#ifndef GENERAL__AT_EXIT_HH
#define GENERAL__AT_EXIT_HH

#include <list>

#include "general/Singleton.hh"

namespace General
{
  ///--------------------------------------------------------------------
  /// \brief  Keep track of cleanup routines at normal program termination
  ///--------------------------------------------------------------------
  class AtExit
  {
    SINGLETON_TS_DECL( AtExit );

  public:
    ///------------------------------------------------------------------
    /// \brief  Type specifier for exit functions.
    ///------------------------------------------------------------------
    typedef void (*ExitFunction)( );

    /// \brief  Add a cleanup function to call when program is exiting.
    static void Append( ExitFunction Function,
			const std::string& Name,
			int Ring );

    static void Cleanup( );

    static bool IsExiting( );

  private:
    ///------------------------------------------------------------------
    /// \brief Stack to keep track of exit functions to call.
    ///
    /// Cannot trust any of the stl objects to be available once exit()
    /// has been called. This structure keeps track of exit routines
    /// that need to be called.
    ///------------------------------------------------------------------
    struct exit_function_node {
      ///----------------------------------------------------------------
      /// Function to be called
      ///----------------------------------------------------------------
      ExitFunction		m_function;

      char*			m_name;

      ///----------------------------------------------------------------
      /// Link to the previous node in the list. If the value of this
      /// object is NULL, then the object is the bottom of the stack.
      ///----------------------------------------------------------------
      exit_function_node*	m_next;

      ///----------------------------------------------------------------
      /// \brief Constructor
      ///
      /// The constructor creates a node which holds the function
      /// to be called and a link to the previous node.
      ///
      /// \param Function The function to called to release resources.
      /// \param Name The name associated with the exit function.
      /// \param Next The address of the next node, NULL otherwise.
      ///----------------------------------------------------------------
      exit_function_node( ExitFunction Function,
			  const std::string& Name,
			  exit_function_node* Next );
    };
    ///------------------------------------------------------------------
    /// \brief Stack to keep track of exit rings.
    ///
    ///------------------------------------------------------------------
    struct exit_ring_node {
      int 			m_ring;
      exit_ring_node*		m_next;
      exit_function_node*	m_exit_function_list;

      ///----------------------------------------------------------------
      /// \brief Constructor
      ///
      /// The constructor creates a node which holds the function
      /// to be called and a link to the previous node.
      ///
      /// \param Function The function to called to release resources.
      /// \param Next The address of the next node, NULL otherwise.
      ///----------------------------------------------------------------
      exit_ring_node( int Ring,
		      exit_ring_node* Next )
	: m_ring( Ring ),
	  m_next( Next ),
	  m_exit_function_list( (exit_function_node*)NULL )
      {
      }
    };

    static exit_ring_node*	m_head;
    static bool			m_is_exiting;

    ///------------------------------------------------------------------
    /// \brief  Destructor
    ///------------------------------------------------------------------
    ~AtExit( );
  };

  ///--------------------------------------------------------------------
  /// Return true if called while exiting, false otherwise.
  ///--------------------------------------------------------------------
  inline bool AtExit::
  IsExiting( )
  {
    return m_is_exiting;
  }
} // namespace - General

#endif /* GENERAL__AT_EXIT_HH */
